from typing import List

from h1st.model.ml_model import MLModel
from h1st.model.model import Model


class Ensemble(MLModel):
    """
    Base Ensemble class to implement various ensemble classes in the future
    even though we currently only have StackEnsemble.
    """

    def __init__(self, ensembler, sub_models: List[Model]):
        """
        :param ensembler: ensembler model, currently it is sklearn's MultiOutputClassifier
            when framework supports StackEnsembleRegressor,
            ensembler could be either MultiOutputClassifier or another specific base model
        :param sub_models: list of h1st.Model participating in the stack ensemble
        """
        self.base_model = ensembler
        self._sub_models = sub_models



import os
import pathlib
from typing import Any, NoReturn
import cloudpickle
from distutils import dir_util
from h1st.model.repository.storage.base import Storage


class LocalStorage(Storage):
    """
    Provide data storage on top of local file system
    """
    def __init__(self, storage_path=None):
        self.storage_path = storage_path

    def get_obj(self, name: str) -> Any:
        """
        Retrieve object value

        :param name: object name
        """
        key = self._to_key(name)
        if not os.path.exists(key):
            raise KeyError(name)

        with open(key, "rb") as f:
            return cloudpickle.load(f)

    def get_bytes(self, name: str) -> bytes:
        """
        Retrieve object value in bytes

        :param name: object name
        """
        key = self._to_key(name)
        if not os.path.exists(key):
            raise KeyError(name)

        with open(key, "rb") as f:
            return f.read()

    def set_obj(self, name: str, value: Any) -> NoReturn:
        """
        Set key value to a python object

        :param name: object name
        :param value: value in python object
        """
        key = self._to_key(name)

        os.makedirs(os.path.dirname(key), mode=0o777, exist_ok=True)
        with open(key, "wb") as f:
            return cloudpickle.dump(value, f)

    def set_bytes(self, name, value):
        """
        Set a key value to a list of bytes

        :param name: object name
        :param value: value in bytes
        """
        key = self._to_key(name)

        os.makedirs(os.path.dirname(key), mode=0o777, exist_ok=True)
        with open(key, "wb") as f:
            return f.write(value)

    def exists(self, name: str) -> bool:
        """
        Return true if object exists in the storage
        """
        key = self._to_key(name)
        return os.path.exists(key)

    def delete(self, name: str) -> NoReturn:
        """
        Delete an object in storage
        """
        key = self._to_key(name)

        if os.path.exists(key):
            os.remove(key)

    def list_keys(self, namespace="") -> list:
        key = namespace.replace("/", "_").replace("..", "__").replace("::", "/")
        if key:
            path = self.storage_path + "/" + key
        else:
            path = self.storage_path

        return list(sorted([p.name for p in pathlib.Path(path).glob('*')]))

    def delete_namespace(self, namespace):
        if not namespace:
            return

        key = namespace.replace("/", "_").replace("..", "__").replace("::", "/")
        path = pathlib.Path(os.path.join(self.storage_path, key))
        if path.exists() and path.is_dir():
            dir_util.remove_tree(str(path))

    def _to_key(self, key):
        # TODO: make sure it is a safe name
        key = key.replace("/", "_").replace("..", "__").replace("::", "/")
        return f"{self.storage_path}/{key}"



import sys
import os
import pathlib
import logging

from h1st.model.repository.model_repository import ModelRepository

class Context:
    def __init__(self):
        pass

    @classmethod
    def init_model_repo(cls, repo_path):
        if not hasattr(ModelRepository, 'MODEL_REPO'):
            setattr(ModelRepository, 'MODEL_REPO', ModelRepository(storage=repo_path))


def init(address=None, MODEL_REPO_PATH=None):
    # TODO: notebook or script only?
    module_path, parent_path = discover_h1st_project()

    # so that project package can be importable immediately
    if parent_path and parent_path not in sys.path:
        sys.path.append(parent_path)

    if MODEL_REPO_PATH:
        Context.init_model_repo(MODEL_REPO_PATH)


def discover_h1st_project(cwd=None) -> tuple:
    # add current dir to cwd
    sys.path.append(os.getcwd())

    cur = pathlib.Path(cwd or os.getcwd())
    last_is_module = None

    module_path = None
    parent_path = None

    while cur != cur.parent:
        # detect the root of the H1ST project structure
        # some project does not have graph.py yet, so we only look for config file for now
        if (cur / "config.py").exists():
            last_is_module = True
            module_path = str(cur)
        elif last_is_module:
            parent_path = str(cur)
            break

        cur = cur.parent

    return module_path, parent_path


_stream_handler = None
def setup_logger():
    global _stream_handler

    name = __name__.split(".")[0]
    logger = logging.getLogger(name)

    if not _stream_handler:
        _stream_handler = logging.StreamHandler()
        _stream_handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(name)s: %(message)s'))

        logger.addHandler(_stream_handler)

    logger.setLevel(logging.INFO)



from typing import Any

from .model import Model


class PredictiveModel(Model):
    """
    Base class for all predictive models.
    """

    def predict(self, input_data: dict) -> dict:
        """
        Implement logic to generate prediction from data

        :params data: data for prediction
        :returns: prediction result as a dict
        """
        return self.process(input_data = input_data)



from typing import Any

from h1st.h1flow.h1step_containable import NodeContainable

from .modelable import Modelable


class Modeler(NodeContainable):
    """
    Base class for H1st Modeler.

    To create your own modeller, inherit `Modeler` class and implement `load_data`, `explore`, `train`,
    `evaluate and `build` accordingly. Please refer to Tutorial for more details how to create a model.
    The framework allows you to persist and load model to the model repository.
    To persist the model, you can call `persist()`, and then `load` to retrieve the model.
    See `persist()` and `load()` document for more detail.

        .. code-block:: python
           :caption: Model Persistence and Loading Example

           import h1st

           class MyModeler(h1st.model.Modeler):
               def train(self, prepared_data):
                   X, y = prepared_data['X'], prepared_data['y']
                   ...
               def build(self):
                   ...

           class MyModel(h1st.model.Model):
               


           my_modeler = MyModeler()
           my_modeler.model_class = MyModel

           my_model = my_modeler.build_model()

           # Persist the model to repo
           my_model.persist('1st_version')

           # Load the model from the repo
           my_model_2 = MyModel()
           my_model_2.load_params('1st_version')
    """
    @property
    def model_class(self) -> Any:
        return getattr(self, "__model_class", None)

    @model_class.setter
    def model_class(self, value):
        setattr(self, "__model_class", value)

    @property
    def stats(self):
        return getattr(self, '__stats__', None)

    @stats.setter
    def stats(self, value) -> dict:
        setattr(self, '__stats__', value)

    @property
    def metrics(self):
        if not hasattr(self, '__metrics__'):
            setattr(self, '__metrics__', {})
        return getattr(self, '__metrics__')

    @metrics.setter
    def metrics(self, value) -> dict:
        setattr(self, '__metrics__', value)
    
    def load_data(self) -> dict:
        """
        Implement logic of load data from data source

        :returns: loaded data
        """

    def explore_data(self, loaded_data: dict) -> None:
        """
        Implement logic to explore data from loaded data
        """

    def evaluate_model(self, prepared_data: dict, model: Modelable) -> dict:
        """
        Implement logic to evaluate the model using the prepared_data
        This function will calculate model metrics and store it into self.metrics

        :param data: loaded data
        """
        if type(model) != self.model_class:
            raise ValueError('The provided model is not a %s' % self.model_class.__name__)
        
        return None
    
    def build_model(self) -> Modelable:
        """
        Implement logic to create the corresponding Model object
        """



from typing import Dict, List, NoReturn
from sklearn.metrics import confusion_matrix
from sklearn.metrics import precision_recall_fscore_support
from sklearn.metrics import accuracy_score
from sklearn.multioutput import MultiOutputClassifier

from h1st.model.ensemble.stack_ensemble import StackEnsemble
from h1st.model.model import Model


class ClassifierStackEnsemble(StackEnsemble):
    """StackEnsemble classifier

    This is the base class for stack ensemble classifiers
    """

    def __init__(self, ensembler: MultiOutputClassifier, sub_models: List[Model], **kwargs):
        super().__init__(ensembler, sub_models, **kwargs)

    def evaluate(self, prepared_data: Dict, metrics: List[str]=None) -> NoReturn:
        """
        Evaluates for the test data
        :param prepared_data: a dictionary {'X_test': ..., 'y_test': ...}
        :param metrics: list of metrics to return and to persist later by the model.
            Default value = ['confusion_matrix', 'precision', 'recall', 'f1', 'support', 'accuracy']

        """
        def add_metric(name, value):
            if name in metrics:
                self.metrics[name] = value

        if not metrics:
            metrics = ['confusion_matrix', 'precision', 'recall', 'f1', 'support', 'accuracy']

        X_test, y_test = prepared_data['X_test'], prepared_data['y_test']
        y_pred = self.predict({'X': X_test})['predictions']
        
        precision, recall, f1, support = precision_recall_fscore_support(y_test, y_pred)

        add_metric('confusion_matrix', confusion_matrix(y_test, y_pred))        
        add_metric('precision', precision)
        add_metric('recall', recall)
        add_metric('f1', f1)
        add_metric('support', support)
        add_metric('accuracy', accuracy_score(y_test, y_pred))



from typing import List

from sklearn.ensemble import RandomForestClassifier

from h1st.model.ensemble.classifier_stack_ensemble import ClassifierStackEnsemble
from h1st.model.model import Model


class RandomForestClassifierStackEnsemble(ClassifierStackEnsemble):
    """
    A ready to use StackEnsemble for classifier with ensembler is a sklearn's MultiOutputClassifier using RandomForestClassifier

    Each sub model must be a subclass of h1.Model and its .predict() method will receive an input data as a dictionary that has 'X' key and numeric value
    and return a dictionary with 'predictions' key and its prediction value

    .. code-block:: python
        :caption: Sub model for a StackEnsemble Example

        class Model1(h1.Model):
            def predict(self, data):
                X = data['X']
                ...
                return {'predictions': }

    .. code-block:: python
        :caption: RandomForestClassifierStackEnsemble usage Example

        class Model2(h1.Model):
                def predict(self, data):
                    X = data['X']
                    ...
                    return {'predictions': }

        class RandomForestClassifierStackEnsemble(ClassifierStackEnsemble):
            def __init__(self):
                super().__init__([
                    Model1().load_params('version_of_model_1'),
                    Model2().load_params('version_of_model_2')
                ])

            def load_data(self,):
                ...
                return loaded_data

            def prep(self, loaded_data):
                ...
                return prepared_data

        m1 = Model1()

        m1.load_prep_train_eval()
        ## Equivalent to
        # loaded_data = m1.load_data()
        # prepared_data = m1.prep(loaded_data)
        # m1.train(prepared_data)
        # m1.evaluate(prepared_data)

        print(m1.metrics)
        m1.persist('version_of_model_1')

        m2 = Model2()
        m2.load_prep_train_eval()
        print(m2.metrics)
        m2.persist('version_of_model_2')

        ensemble = RandomForestClassifierStackEnsemble(
            [Model1().load_params'version_of_model_1'),
             Model2().load_params('version_of_model_2')])
        ensemble.load_prep_train_eval()
        print(ensemble.metrics)
        ensemble.persist('version_of_model_ensemble')
        ensemble.predict(...)
    """

    def __init__(self, sub_models: List[Model], **kwargs):
        super().__init__(
            # MultiOutputClassifier(RandomForestClassifier(n_jobs=-1, max_depth=4, random_state=42)),
            RandomForestClassifier(n_jobs=-1, max_depth=4, random_state=42),
            sub_models,
            **kwargs
        )



from h1st.model.predictive_model import PredictiveModel

from .student import Student, StudentModeler
from .teacher import Teacher, TeacherModeler
from .ensemble import Ensemble

"""
Oracle architecture:

@startuml
allowmixing

Component Oracle #EEE {
	Class Teacher
	Class Student
	Class Ensemble
}

Actor "AI Engineer" as User

User .down.> TeacherModeler : uses
User .down.> StudentModeler : uses

TeacherModeler -down-> Teacher : builds
StudentModeler -down-> Student : builds
Teacher .right.> Student : teaches



Teacher -down-> Ensemble : trains
Student -down-> Ensemble : trains

Note as N1  #green
<size:16><color:white>Construction Phase</color></size>
end Note
@enduml
"""

"""
@startuml
allowmixing

Component Oracle #EEE {
    Class Teacher <<RuleBasedModel>>
    Class Student <<ML Generalizer>>
	Class Ensemble
}

Database Data
Circle Prediction

Data -down-> Teacher
Data -down-> Student

Teacher -down-> Ensemble : votes
Student -down-> Ensemble : votes

Ensemble -down-> Prediction

Note as N1  #green
<size:16><color:white>Execution Phase</color></size>
end Note
@enduml
"""

class Oracle(PredictiveModel):
    """
    Teacher Model in Oracle framework
    """
    pass



class Describer:
    """
    A Describer is one that stores all relevant details about the objects (e.g., `Models`, `Graphs`).
    This is important to have because if and when a constituent asks about the `Model` or `Graph`,
    the Describer will provide all pertenant details. For example, if a Regulator identifies that the model
    is indicating bias in its decision making, she may asks for details such as `when was the model trained',
    `by whom' and 'on what data was the model trained'
    """

    def __init__(self, h1stmodel):
        self.shap_describer = None
        self.data_description(h1stmodel)
        self.model_description(h1stmodel)
        
    def data_description(self, h1stmodel):
        self.h1stmodel = h1stmodel
        _dict = {}
        data = h1stmodel.prepared_data
        _dict["data_set_name"] = h1stmodel.dataset_name
        _dict["data_set_description"] = h1stmodel.dataset_description
        _dict["label_column"] = h1stmodel.label_column
        _dict["features"] = list(data["train_df"].columns)
        _dict["number_of_features"] = len(_dict["features"])
        _dict["number_of_rows"] = data["train_df"].shape[0]
        _dict["statistics"] = data["train_df"].describe()
        self.data_describer = _dict

    def model_description(self, h1stmodel):
        _dict = {}
        model = h1stmodel._native_model
        _dict["model_name"] = str(type(model).__name__)
        _dict["model_params"] = model.get_params()
        _dict["model_metrics"] = h1stmodel.metrics
        self.model_describer = _dict

    def generate_report(self, constituency, aspect):
        pass


import os
import pandas as pd
from typing import Union, Optional, Callable, List, NoReturn, Any, Dict, Tuple

from h1st.exceptions.exception import GraphException
from h1st.model.model import Model
from h1st.h1flow.h1step_containable import NodeContainable
# from h1st.schema import SchemaValidator

class Node:
    """
    Base class for h1st Node
    """

    def __init__(self, containable: NodeContainable = None, id: str = None):
        """
        :param containable: instance of subclass of NodeContainable to attach to the node
        :param id: the node's id
        """
        if containable:
            containable._node = self
        
        self._id = id
        self._containable = containable
        self._graph = None
        self._edges = []        

        self._transform_input = None
        self._transform_output = None

        # viz attribute
        self.rank = None

    @property
    def id(self) -> str:
        return self._id

    @property
    def edges(self) -> List[Tuple['Node', str]]:
        """
        :returns: list of tuple(next_node, edge_label) which next_node is connected from this node
            edge_label = 'yes'/'no' in case of condition nodes
            edge_label = None in case of normal nodes            
        """
        return self._edges

    @property
    def graph(self) -> 'Graph':
        return self._graph

    @graph.setter
    def graph(self, value):
        if self.graph:
            raise GraphException('This node belongs to another graph already')

        self._graph = value

    @property
    def transform_input(self) -> Callable:
        return self._transform_input

    @transform_input.setter
    def transform_input(self, value):
        """
        Transforms input of a node before executing the node

        .. code-block:: python
            :caption: Example of transform_input

            import h1st as hf

            class MyModel(h1.Model):
                pass

            class MyGraph(h1.Graph)
                def __init__(self):
                    self.start()
                        .add(MyModel(), id="my_model")
                        .end()

                    self.nodes.my_model.transform_input = self._transform_my_model_input

                def _transform_my_model_input(self, inputs):
                    inputs['xxx'] = ...
                    return inputs
        """
        self._transform_input = value

    @property
    def transform_output(self) -> Callable:
        return self._transform_output

    @transform_output.setter
    def transform_output(self, value):
        """
        Transforms output of a node after executing the node

        .. code-block:: python
            :caption: Example of transform_output

            import h1st as hf

            class MyModel(h1.Model):
                pass

            class MyGraph(h1.Graph)
                def __init__(self):
                    self.start()
                        .add(MyModel())
                        .end()

                    self.nodes.end.transform_output = self._transform_end_output

                def _transform_end_output(self, inputs):
                    return {
                        'result': inputs['xxx']
                    }
        """        
        self._transform_output = value

    def add(
        self,
        node: Union['Node', NodeContainable, None] = None,
        yes: Union['Node', NodeContainable, None] = None,
        no: Union['Node', NodeContainable, None] = None,
        id: str = None
     ) -> Union['Node', List['Node']]:
        """
        The bridge function to add nodes to a graph. This will invoke the Graph.add() function and 
        will then connect this node to newly added nodes.
        """
        return self._graph._add_and_connect(node, yes, no, id, self)

    def _execute(self, command: Optional[str], inputs: Dict[str, Any], state: Dict) -> Dict:
        """
        The execution of graph will be executed recursively. The upstream node will invoke the down stream nodes to be executed.
        If it is the start node, this function will be invoked by the graph.
        The containable.call() will be invoked if this node contains a NodeContainable object. Otherwise, its call() function will be invoked.

        :param command: for this node to know which flow (predict, train, ...) the graph is running
        :param inputs: the input data to execute the node. During the graph execution, output of all executed nodes will be merged into inputs
        :param state: executing state
        """
        # transform input
        if callable(self.transform_input):
            inputs = self.transform_input(inputs)

        # execute        
        node_output = self.call(command, inputs)

        # transform output
        if self.id != "end" and callable(self.transform_output):
            node_output = self.transform_output({**inputs, **node_output})

        # validate output
        self._validate_output(node_output)

        # state = state or {}
        if node_output:
            state.update(node_output)
        else:
            node_output = {}

        # recursively executing downstream nodes
        for edge in self.edges:
            edge_data = self._get_edge_data(edge, node_output)

            # data is available to execute the next node
            if edge_data is not None:                
                next_node = edge[0]
                next_inputs = {**inputs, **edge_data}
                next_node._execute(command, next_inputs, state)

        return {**node_output, **state}

    def call(self, command: Optional[str], inputs: Dict[str, Any]) -> Dict:
        """
        Subclass may need to override this function to perform the execution depending the type of node.
        This function is invoked by the framework and user will never need to call it.
        """
        if self._containable:
            return self._containable.call(command, inputs)

        return {}

    def to_dot_node(self, visitor):
        """Subclass will need to implement this function to construct and return the graphviz compatible node"""

    def validate_output(self, input_data: dict = None, schema=None, command='predict'):
        """
        Invokes the call function the node with given input data, then verifies if output of the node conforming with the output schema of this node.
        Framework will look up the output schema for this node in the schemas object loaded by the graph from schemas.py using id of this node.

        :param inputs: input data to invoke the call function of the node
        :param schema: provided schema to validate with output
        :param command: the command param to invoke the call function of the node
        """
        if self._containable:
            return self._containable.validate_node_output(input_data, schema)
        else:
            output = self.call(command, input_data)
            return SchemaValidator().validate(output, schema)

    def _get_edge_data(self, edge, node_output):
        """Gets data from node's output to pass to the next node"""
        return node_output

    def _validate_output(self, node_output) -> bool:
        return True


class Action(Node):
    """
    @TODO: use .add(yes = (Model2(), "model2")) instead ?

    H1st non-conditional node. It is used to add a NodeContainable instance with an ID to yes/no branch of a conditional node

    .. code-block:: python
        :caption: Add nodes for yes/no branch of a conditional node

        import h1st.core as h1

        class MyGraph(h1.Graph)
            def __init__(self):
                yes, no = self.start()
                    .add(h1.Decision(Model1()))
                    .add(
                        yes = h1.Action(Model2(), id="model2"), # with an id
                        no = Model3()                           # without an id
                    )
                
                yes.add(Model4())
                no.add(Model5())

                self.end()
    """

    def to_dot_node(self, visitor):
        """Constructs and returns the graphviz compatible node"""
        return visitor.render_dot_action_node(self)


class NoOp(Action):
    """A do-nothing action."""
    def call(self, command, inputs):
        pass


class Decision(Action):
    """
    H1st conditional node

    .. code-block:: python
        :caption: Graph with conditional node

        import h1st.core as h1

        class MyGraph(h1.Graph)
            def __init__(self):
                yes, no = self.start()
                    .add(h1.Decision(Model1()))
                    .add(
                        yes = Model2()
                        no = Model3()
                    )

                yes.add(Model4())
                no.add(Model5())
                
                self.end()

    .. code-block:: python
        :caption: Graph with conditional node using custom result_field and decision_field

        import h1st.core as h1

        class Model1(h1.Model):
            def predict(data):
                return {
                    'predictions': [
                        {gx: 10, gy: 20, label: True},
                        {gx: 11, gy: 21, label: True},
                        {gx: 12, gy: 22, label: False},
                    ]
                }

        class MyGraph(h1.Graph)
            def __init__(self):
                yes, no = self.start()
                    .add(h1.Decision(Model1(), result_field='predictions', decision_field='label'))
                    .add(
                        yes = Model2()
                        no = Model3()
                    )

                yes.add(Model4())
                no.add(Model5())
                
                self.end()
    """

    def __init__(self, containable: NodeContainable = None, id: str = None, result_field='results', decision_field='prediction'):
        """
        :param containable: instance of subclass of NodeContainable to attach to the node
        :param id: the node's id
        :param result_field: the key to extract the data collection from dictionary output of a conditional node
        :param decision_field: the field name to decide which item of the collection belongs to yes branch or no branch
        """
        super().__init__(containable, id)
        self._result_field = result_field
        self._decision_field = decision_field

    def to_dot_node(self, visitor):
        """Constructs and returns the graphviz compatible node"""
        return visitor.render_dot_decision_node(self)

    def _get_edge_data(self, edge, node_output):
        """splits data for yes/no path from the node's output to pass to the next node"""
        result_field = self._result_field if self._result_field in node_output else next(iter(node_output))
        results = node_output[result_field]

        decision_field = self._decision_field
        is_yes_edge = edge[1] == 'yes'

        if isinstance(results, pd.DataFrame):
            data = results[results[decision_field] == is_yes_edge]
        else:
            data = [item for item in results if item[decision_field] == is_yes_edge]

        return {result_field: data} if data is not None and len(data) > 0 else None

    def _validate_output(self, node_output) -> bool:
        """
        This will ensure the result's structure is valid for decision node.
        
        node_output must be a dictionary containing 'results' key and each item will have a field whose name = 'prediction'
        with bool value to decide whether the item belongs to yes or no branch
            { 
                'results': [{ 'prediction': True/False, ...}],
                'other_key': ...,
            }

        or a dictionary containing only one key
            {
                'your_key': [{ 'prediction': True/False, ...}]
            }
        """
        if not isinstance(node_output, dict) or ((self._result_field not in node_output) and len(node_output.keys()) != 1):
            raise GraphException(f'output of {type(self._containable)} must be a dict containing "results" field or only one key')
        
        return True


from typing import Any

from .predictive_model import PredictiveModel


class MLModel(PredictiveModel):
    """
    Base class for H1st Model.

    To create your own Machine Learning model, inherit `MLModel` class and implement `prepare_data`, 
    `train` and `predict` accordingly. Please refer to Tutorial for more details how to create a model.

    The framework allows you to persist and load model to the model repository.
    To persist the model, you can call `persist()`, and then `load` to retrieve the model.
    See `persist()` and `load_params()` document for more detail.

        .. code-block:: python
           :caption: Model Persistence and Loading Example

           import h1st.core as h1
           from sklearn.datasets import load_iris
           from sklearn.linear_model import LogisticRegression

           class MyModel(h1.Model):
               def __init__(self):
                   super().__init__()
                   self.model = LogisticRegression(random_state=0)

               def train(self, prepared_data):
                   X, y = prepared_data['X'], prepared_data['y']
                   self.model.fit(X, y)

           my_model = MyModel()
           X, y = load_iris(return_X_y=True)
           prepared_data = {'X': X, 'y': y}
           my_model.train(prepared_data)
           # Persist the model to repo
           my_model.persist('1st_version')

           # Load the model from the repo
           my_model_2 = MyModel()
           my_model_2.load_params('1st_version')
    """

    @property
    def base_model(self) -> Any:
        return getattr(self, "__base_model", None)

    @base_model.setter
    def base_model(self, value):
        setattr(self, "__base_model", value)



import os
import tarfile
import tempfile
import logging
import importlib
from distutils import dir_util

import yaml
import ulid

from h1st.model.repository.storage.s3 import S3Storage
from h1st.model.repository.storage.local import LocalStorage

SEP = "::"
logger = logging.getLogger(__name__)


class ModelSerDe:
    STATS_PATH = 'stats.joblib'
    METRICS_PATH = 'metrics.joblib'
    METAINFO_FILE = 'METAINFO.yaml'

    def _serialize_dict(self, d, path, dict_file):
        import joblib
        joblib.dump(d, path + '/%s' % dict_file)

    def _deserialize_dict(self, path, dict_file):
        import joblib
        return joblib.load(path + '/%s' % dict_file)

    def _get_model_type(self, model):
        import tensorflow
        import sklearn

        if isinstance(model, sklearn.base.BaseEstimator):
            return 'sklearn'
        if isinstance(model, tensorflow.keras.Model):
            return 'tensorflow-keras'
        if model is None:
            return 'custom'

    def _serialize_single_model(self, model, path, model_name='model'):
        model_type = self._get_model_type(model)

        if model_type == 'sklearn':
            # This is a sklearn model
            import joblib
            model_path = '%s.joblib' % model_name
            joblib.dump(model, path + '/%s' % model_path)
        elif model_type == 'tensorflow-keras':
            model_path = model_name
            os.makedirs(path + '/%s' % model_path, exist_ok=True)
            model.save_weights(path + '/%s/weights' % model_path)
        elif model_type == 'custom':
            model_path = model_name  # XXX
        else:
            raise ValueError('Unsupported model!!!')

        return model_type, model_path

    def _deserialize_single_model(self, model, path, model_type, model_path):
        if model_type == 'sklearn':
            # This is a sklearn model
            import joblib
            model = joblib.load(path + '/%s' % model_path)
            # print(str(type(model)))
        elif model_type == 'tensorflow-keras':
            model.load_weights(path + '/%s/weights' % model_path).expect_partial()
        elif model_type == 'custom':
            model = None

        return model

    def serialize(self, model, path):
        """
        Serialize a H1ST model's model property to disk.

        :param model: H1ST Model
        :param path: path to save models to
        """
        from h1st.model.ml_model import MLModel
        from h1st.model.rule_based_model import RuleBasedModel

        meta_info = {}

        if model.metrics:
            logger.info('Saving metrics property...')
            meta_info['metrics'] = self.METRICS_PATH
            self._serialize_dict(model.metrics, path, self.METRICS_PATH)

        if model.stats is not None:
            logger.info('Saving stats property...')
            meta_info['stats'] = self.STATS_PATH
            self._serialize_dict(model.stats, path, self.STATS_PATH)

        if isinstance(model, MLModel):
            if model.base_model:
                logger.info('Saving model property...')
                if type(model.base_model) == list:
                    meta_info['models'] = []
                    for i, model in enumerate(model.base_model):
                        model_type, model_path = self._serialize_single_model(model, path, 'model_%d' % i)
                        meta_info['models'].append({'model_type': model_type, 'model_path': model_path})
                elif type(model.base_model) == dict:
                    meta_info['models'] = {}
                    for k, model in model.base_model.items():
                        model_type, model_path = self._serialize_single_model(model, path, 'model_%s' % k)
                        meta_info['models'][k] = {'model_type': model_type, 'model_path': model_path}
                else:
                    # this is a single model
                    model_type, model_path = self._serialize_single_model(model.base_model, path)
                    meta_info['models'] = [{'model_type': model_type, 'model_path': model_path}]
            else:
                logger.error('.base_model was not assigned.')

        elif not isinstance(model, RuleBasedModel):
            pass
        elif hasattr(model, 'base_model'):            
            logger.warning('Your .base_model will not be persisted. If you want to persist your .base_model, \
                            must inherit from h1st.MLModel instead of h1st.Model.')

        if len(meta_info) == 0:
            logger.info('Model persistence currently supports only stats, model and metrics properties.')
            logger.info('Make sure you store stastistic in stats property, models in model property and model metrics in metrics one.')

        with open(os.path.join(path, self.METAINFO_FILE), 'w') as file:
            yaml.dump(meta_info, file)

    def deserialize(self, model, path):
        """
        Populate a H1ST model's model property with saved atomic models.

        :param model: H1ST Model
        :param path: path to model folder
        """
        # Read METAINFO
        with open(os.path.join(path, self.METAINFO_FILE), 'r') as file:
            meta_info = yaml.load(file, Loader=yaml.Loader)

        if 'metrics' in meta_info.keys():
            model.metrics = self._deserialize_dict(path, self.METRICS_PATH)

        if 'stats' in meta_info.keys():
            model.stats = self._deserialize_dict(path, self.STATS_PATH)

        if 'models' in meta_info.keys():
            model_infos = meta_info['models']
            org_model = model.base_model  # original model object from Model class
            if type(model_infos) == list:
                if len(model_infos) == 1:
                    # Single model
                    model_info = model_infos[0]
                    # print(model_info)
                    model_type = model_info['model_type']
                    model_path = model_info['model_path']
                    model.base_model = self._deserialize_single_model(org_model, path, model_type, model_path)
                else:
                    # A list of models
                    model.base_model = []
                    org_model = org_model or [None for _ in range(len(model_infos))]
                    for i, model_info in enumerate(model_infos):
                        model_type = model_info['model_type']
                        model_path = model_info['model_path']
                        model.base_model.append(self._deserialize_single_model(org_model[i], path, model_type, model_path))

            elif type(model_infos) == dict:
                # A dict of models
                model.base_model = {}
                org_model = org_model or {k: None for k in model_infos.keys()}
                for model_name, model_info in model_infos.items():
                    model_type = model_info['model_type']
                    model_path = model_info['model_path']
                    model.base_model[model_name] = self._deserialize_single_model(org_model[model_name], path, model_type, model_path)
            else:
                raise ValueError('Not a valid H1ST Model METAINFO file!')


class ModelRepository:
    """
    Model repository allows user to persist and load model to different storage system.

    Model repository uses ``ModelSerDer`` to serialize a model into a temporary folder
    and then create a tar archive to store on storage. For loading, the repo retrieve
    the tar archive from the storage and then extract to temporary folder for restoring
    the model object.
    """

    _NAMESPACE = "_models"

    _DEFAULT_STORAGE = S3Storage

    def __init__(self, storage=None):
        if isinstance(storage, str) and "s3://" in storage:
            storage = storage.replace("s3://", "").strip("/") + "/"
            bucket, prefix = storage.split("/", 1)
            storage = S3Storage(bucket, prefix.strip("/"))
            self._NAMESPACE = ""
        elif isinstance(storage, str):  # local folder
            storage = LocalStorage(storage)
            self._NAMESPACE = ""

        self._storage = storage or ModelRepository._DEFAULT_STORAGE()
        self._serder = ModelSerDe()

    def persist(self, model, version=None):
        """
        Save a model to the model repository

        :param model: target model
        :param version: version name, leave blank for autogeneration
        :returns: model version
        """
        # assert isinstance(model, Model)
        # TODO: use version format: v_20200714-1203
        version = version or str(ulid.new())

        try:
            # serialize a model to a temporary folder and then clean up later
            tmpdir = tempfile.mkdtemp()
            serialized_dir = os.path.join(tmpdir, 'serialized')
            tar_file = os.path.join(tmpdir, 'model.tar')
            os.makedirs(serialized_dir)

            self._serder.serialize(model, serialized_dir)
            _tar_create(tar_file, serialized_dir)

            with open(tar_file, mode='rb') as f:
                self._storage.set_bytes(
                    self._get_key(model, version),
                    f.read(),
                )

                self._storage.set_obj(
                    self._get_key(model, 'latest'),
                    version,
                )

                model.version = version
        finally:
            dir_util.remove_tree(tmpdir)

        return version

    def load(self, model, version=None):
        """
        Restore the model from the model repository

        :param model: target model
        :param version: version name, leave blank to load the latest version
        """
        # assert isinstance(model, Model)
        if version is None:
            version = self._storage.get_obj(self._get_key(model, 'latest'))

        logger.info('Loading version %s ....' % version)

        try:
            tmpdir = tempfile.mkdtemp()
            serialized_dir = os.path.join(tmpdir, 'serialized')
            tar_file = os.path.join(tmpdir, 'model.tar')
            os.makedirs(serialized_dir)

            with open(tar_file, 'wb') as f:
                f.write(self._storage.get_bytes(
                    self._get_key(model, version)
                ))

            _tar_extract(tar_file, serialized_dir)
            self._serder.deserialize(model, serialized_dir)
            model.version = version
        finally:
            # We get error from Tensorflow telling that it could not find the folder
            # Unsuccessful TensorSliceReader constructor: Failed to get matching files on 
            # /var/folders/wb/40304xlx477cfjzbk386l2gr0000gn/T/tmpwcrvm2e2/model/weights: 
            # Not found: /var/folders/wb/40304xlx477cfjzbk386l2gr0000gn/T/tmpwcrvm2e2/model; No such file or directory [Op:RestoreV2]
            #
            # dir_util.remove_tree(tmpdir)

            # instead, register the function to clean it up when the interpreter quits
            import atexit

            def clean_tmpdir(tmpdir):
                # print('Clean up %s'  % tmpdir)
                dir_util.remove_tree(tmpdir)

            atexit.register(clean_tmpdir, tmpdir=tmpdir)

    def delete(self, model, version):
        """
        Delete a model from model repository

        :param model: model instance or the model class
        :param version: target version
        """
        # assert isinstance(model, Model) or isinstance(model, type)
        assert version != 'latest'  # magic key
        self._storage.delete(self._get_key(model, version))

    def download(self, model, version, path):
        """
        Download a model archive to local disk

        :param model: model instance or model class
        :param version: version name
        :param path: target folder to extract the model archive
        """
        with tempfile.NamedTemporaryFile(mode="wb") as f:
            f.write(self._storage.get_bytes(
                self._get_key(model, version)
            ))
            f.flush()
            f.seek(0)

            _tar_extract(f.name, path)

        return path

    # TODO: list all versions

    def _get_key(self, model, version):
        model_class = model if isinstance(model, type) else model.__class__
        model_name = model_class.__module__ + '.' + model_class.__name__

        key = f"{model_name}{SEP}{version}"

        if self._NAMESPACE:
            key = f"{self._NAMESPACE}{SEP}{key}"

        return key

    @classmethod
    def get_model_repo(cls, ref=None):
        """
        Retrieve the default model repository for the project

        :param ref: target model
        :returns: Model repository instance
        """
        if not hasattr(cls, 'MODEL_REPO'):   # global ModelRepository.MODEL_REPO
            repo_path = None
            if ref is not None:
                # root module
                root_module_name = ''
                
                # find the first folder containing config.py to get MODEL_REPO_PATH
                for sub in ref.__class__.__module__.split('.'):
                    root_module_name = sub if not root_module_name else root_module_name + '.' + sub

                    try:
                        module = importlib.import_module(root_module_name + ".config")
                        repo_path = getattr(module, 'MODEL_REPO_PATH', None)
                        break
                    except ModuleNotFoundError:
                        repo_path = None

            # in the new structure, the config file may be at root folder
            if not repo_path:
                try:
                    import config
                    repo_path = config.MODEL_REPO_PATH
                except ModuleNotFoundError:
                    repo_path = None

            if not repo_path:
                repo_path = os.environ.get('H1ST_MODEL_REPO_PATH', '')

            if not repo_path:
                raise RuntimeError('Please set MODEL_REPO_PATH in config.py')

            setattr(cls, 'MODEL_REPO', ModelRepository(storage=repo_path))

        return getattr(cls, 'MODEL_REPO')


def _tar_create(target, source):
    """
    Helper function to create a tar archive
    """
    with tarfile.open(target, "w:gz") as tf:
        tf.add(source, arcname='', recursive=True)

    return target


def _tar_extract(source, target):
    """
    Helper function to extract a tar archive
    """
    with tarfile.open(source) as tf:
        tf.extractall(target)



from h1st.model.ml_model import MLModel
from h1st.model.ml_modeler import MLModeler


class Student(MLModel):
    """
    Knowledge Generalization Model
    """
    pass


class StudentModeler(MLModeler):
    """
    Knowledge Generalization Modeler
    """
    pass



from h1st.model import Model as H1stModel

# __TODO__: what is this? Does it belong here at all?


class GoogleCloudTranslateModel(H1stModel):
    def predict(self, text, src_lang='en', target_lang='es'):
        from google.cloud import translate_v2 as translate

        translate_client = translate.Client()

        result = translate_client.translate(text,
                                            source_language=src_lang,
                                            target_language=target_lang)

        return result['translatedText']



from typing import Dict, List, Any

import numpy as np
from sklearn.preprocessing import RobustScaler

from h1st.exceptions.exception import ModelException
from h1st.model.ensemble.ensemble import Ensemble
from h1st.model.model import Model


class StackEnsemble(Ensemble):
    """
    Base StackEnsemble class to implement StackEnsemble classifiers or regressors.
    """

    def __init__(self, ensembler, sub_models: List[Model], **kwargs):
        """
        :param **kwargs: StackEnsemble use h1st models as submodels and the predict function of
        h1st model receives a dictionary and returns a dictionary. users can set the key of these
        dictionaries with the following keyword arguments.

            **submodel_predict_input_key (str): the default value of submodel_input_key is 'X' \n
            **submodel_predict_output_key (str): the default value of submodel_output_key is 'predictions'
        """
        super().__init__(ensembler, sub_models)
        self._submodel_predict_input_key = kwargs.get('submodel_predict_input_key', 'X')
        self._submodel_predict_output_key = kwargs.get('submodel_predict_output_key', 'predictions')

    def _preprocess(self, X: Any) -> Any:
        """
        Predicts all sub models then merge input raw data with predictions for ensembler's training or prediction
        :param X: raw input data
        """
        if isinstance(X, list):
            predictions = [self._get_submodels_prediction(x) for x in X]
            return np.vstack(predictions)
        else:
            return self._get_submodels_prediction(X)

    def _get_submodels_prediction(self, X: Any) -> Any:
        preds = []
        for m in self._sub_models:
            pred = m.predict({self._submodel_predict_input_key: X})
            output_key = pred.get(self._submodel_predict_output_key)
            if output_key is not None:
                preds.append(output_key)
        return np.hstack(preds)

    def train(self, prepared_data: Dict):
        """
        Trains ensembler with input is merging raw data 'X_train' and predictions of sub models, with targets 'y_train'
        'X_train' is supposed to be numeric.
        :param prepared_data: output of prep() implemented in subclass of StackEnsemble, its structure must be this dictionary
            {
                'X_train': ...,
                'X_test': ...,
                'y_train': ...,
                'y_test': ...
            }
        """
        X_train, y_train = (prepared_data['X_train'], prepared_data['y_train'])
        X_train = self._preprocess(X_train)
        scaler = RobustScaler(quantile_range=(5.0, 95.0), with_centering=False)
        self.stats = scaler.fit(X_train)
        X_train = self.stats.transform(X_train)

        self.base_model.fit(X_train, y_train)

    def predict(self, data: Dict) -> Dict:
        """
        Returns predictions of ensembler model for input raw data combining with predictions from sub models

        This method will raise ModelException if ensembler model has not been trained.

        :param data: must be a dictionary {'X': ...}, with X is raw data (numeric only)
        :return:
            output will be a dictionary {'predictions': ....}
        """
        if not self.stats:
            raise ModelException('This model has not been trained')

        X = self._preprocess(data['X'])
        X = self.stats.transform(X)
        return {'predictions': self.base_model.predict(X)}



from typing import Any, NoReturn

from h1st.h1flow.h1step_containable import NodeContainable
from h1st.trust.trustable import Trustable

from .repository.model_repository import ModelRepository
from .modeler import Modelable


class Model(NodeContainable, Trustable, Modelable):
    """
    Base class for H1st Model.

    To create your own model, inherit `Model` class and implement `process` accordingly.
    Please refer to Tutorial for more details how to create a model.

    The framework allows you to persist and load model to the model repository.
    To persist the model, you can call `persist()`, and then `load` to retrieve the model.
    See `persist()` and `load()` document for more detail.

        .. code-block:: python
           :caption: Model Persistence and Loading Example

           import h1st

           class MyModeler(h1st.model.Modeler):
               def train(self, data):
                   X, y = prepared_data['X'], prepared_data['y']
                   ...
               def build(self):
                   ...

           class MyModel(h1st.model.Model):
               


           my_modeler = MyModeler()
           my_modeler.model_class = MyModel

           my_model = my_modeler.build_model()

           # Persist the model to repo
           my_model.persist('1st_version')

           # Load the model from the repo
           my_model_2 = MyModel()
           my_model_2.load_params('1st_version')
    """

    ## TODO: Need a better naming and the definition of the property
    @property
    def stats(self):
        return getattr(self, '__stats__', None)

    @stats.setter
    def stats(self, value) -> dict:
        setattr(self, '__stats__', value)

    @property
    def metrics(self):
        if not hasattr(self, '__metrics__'):
            setattr(self, '__metrics__', {})
        return getattr(self, '__metrics__')

    @metrics.setter
    def metrics(self, value) -> dict:
        setattr(self, '__metrics__', value)

    def persist(self, version=None) -> None:
        """
        Persist this model's properties to the ModelRepository. Currently, only `stats`, `metrics`, `model` properties are supported.

        `model` property could be single model, list or dict of models
        Currently, only sklearn and tensorflow-keras are supported.

        :param version: model version, leave blank for autogeneration
        :returns: model version
        """
        repo = ModelRepository.get_model_repo(self)
        return repo.persist(model=self, version=version)

    def load_params(self, version: str = None) -> None:
        """
        Load parameters from the specified `version` from the ModelRepository.
        Leave version blank to load latest version.
        """
        repo = ModelRepository.get_model_repo(self)
        repo.load(model=self, version=version)

        return self
        
    def process(self, input_data: dict) -> Any:
        """
        Implement logic to process data

        :params data: data to process
        :returns: processing result as Any
        """
        # not raise NotImplementedError so the initial model created by integrator will just work 
        return input_data



class Decision:
    """
    A Decision is one that stores all relevant details pertaining to the decision
    that was made by the Model or Graph, the dataset that was used and the Model
    or Graph that was used to make the decision. For example, `what data was used
    to train the model`, `how many features were used to train the model`
    and `tell me how does the model make decisions`
    """

    def data_description(self, data):
        pass

    def model_description(self, model):
        pass

    def decision_explainer(self, decision):
        pass



import os
from math import inf
from types import SimpleNamespace
from graphviz import Digraph, ExecutableNotFound, Graph
from h1st.h1flow.h1step import Decision

__theme__ = {
    'arrow_size': '0.5',
    'default_shape': 'rectangle',
    'default_style': 'rounded, filled',
    'edge_color': '#607894',
    'fill_color': '#7890AC',
    'font_name': 'helvetica',
    'font_size': '9',
    'text_color': '#7890AC',
}

theme = SimpleNamespace(**__theme__)

class DotGraphVisualizer:
    def __init__(self, graph: 'Graph'):
        self.graph = graph
        self.dot_graph = Digraph()
        self.visitor = GraphVisitor()
        self.nodes = []
        self.edges = []
        self._subgraphs = {}

    def render_dot_nodes(self):
        for rank, n in enumerate(self.graph.nodes.__dict__.values(), start=1):

            if n.rank is None:
                n.rank = str(rank)

            new_node = self.render_dot_node(n)
            self.nodes.append(new_node)
            self.clusterize_node(new_node)

            #special handling for Decision node
            edge_constraint = 'true'
            compass_point = None

            if isinstance(n, Decision):
                edge_constraint = 'false'
                compass_point = 'nw'
            else:
                edge_constraint = 'true'
                compass_point = None

            # '''point to end if no outgoing node is available'''
            # if len(n.edges) == 0:
            #     self.edges.append({
            #         'from': new_node['name'],
            #         'to': 'end', 'label': '',
            #         'constraint': edge_constraint,
            #         'compass_point': compass_point,
            #     })

            #connect edges
            seen = set()
            for ns in n.edges:
                if ns[0].rank is None:
                    ns[0].rank = str(rank + 1)

                connected_node = self.render_dot_node(ns[0])

                # self.nodes.append(connected_node)

                if (new_node['name'], connected_node['name']) in seen:
                    continue

                seen.add((new_node['name'], connected_node['name']))
                seen.add((connected_node['name'], new_node['name']))

                self.clusterize_node(connected_node)

                self.edges.append({
                    'from': new_node['name'], 
                    'to': connected_node['name'],
                    'label': ns[1],
                    'constraint': edge_constraint,
                    'compass_point': compass_point
                })

    def clusterize_node(self, node):
        if self._subgraphs.get(node['rank']) is None:
            self._subgraphs[node['rank']] = []

        self._subgraphs[node['rank']].append(node)

    def render_dot_node(self, node):
        return node.to_dot_node(self.visitor)

    def to_dot(self):
        self.dot_graph = Digraph()
        self.visitor = GraphVisitor()
        self.nodes = []
        self.edges = []
        self._subgraphs = {}

        mg = self.dot_graph

        mg.attr(
            'graph',
            compound='True',
            fontname=theme.font_name,
            fontsize=theme.font_size,
            splines="ortho",
            rankdir="LR"
        )

        mg.attr('node',
            color=theme.text_color,
            fillcolor=theme.fill_color,
            fontname=theme.font_name,
            fontsize=theme.font_size,
            shape=theme.default_shape,
            style=theme.default_style
        )

        mg.attr('edge',
            color=theme.edge_color,
            arrowsize=theme.arrow_size,
            fontname=theme.font_name,
            fontsize=theme.font_size,
            rank='same'
        )

        # mg.node(name='start', label='Start', shape='circle')

        self.render_dot_nodes()

        for i, rank in enumerate(self._subgraphs):
            cluster = self._subgraphs[rank]

            with mg.subgraph() as sg:
                sg.attr(rank='same')

                for subnode in cluster:
                    sg.node(**subnode)

        # for node in self.nodes:
        #     print(node)
        #     mg.node(**node)

        seen = set()  # XXX
        for edge in self.edges:
            if (edge['from'], edge['to']) in seen:
                continue

            seen.add((edge['from'], edge['to']))
            seen.add((edge['to'], edge['from']))

            mg.edge(
                edge['from'], edge['to'],
                label=edge['label']
            )

        mg.node(self.nodes[0]['name'], shape='circle')
        mg.node(self.nodes[-1]['name'], shape='circle')
        # mg.edge('start', self.nodes[0]['name'])

        # mg.subgraph(g)
        return mg

    def render_topology(self, target_file):
        dot = self.to_dot()

        try:
            basename = os.path.basename(target_file)

            if "." in basename:
                bits = basename.split(".")

                ext = bits[-1]
                basename = ".".join(bits[:-1])
            else:
                ext = "png"

            dot.render(
                filename=basename,
                format=ext,
                directory=os.path.dirname(target_file),
                cleanup=True,
            )
        except ExecutableNotFound:
            raise EngineNotAvailableException()

        return dot


    def _repr_svg_(self):
        dot = self.to_dot()
        return dot.pipe(format='svg').decode('utf8')


class EngineNotAvailableException(Exception):
    pass


class GraphVisitor:
    def render_node_label(self, node, extra_label=""):
        # print(node.callable)
        # print(type(node.callable).__name__)
        # return "{}\n {}".format(type(node).__name__, node.id)
        return node.id

    def render_node_name(self, node):
        return "{}\n {}".format(type(node).__name__, id(node))

    def render_dot_decision_node(self, node):
        label = self.render_node_label(node)
        node_name = self.render_node_name(node)
        return dict(name=node_name, label=label, shape="diamond", style="filled", rank=node.rank)

    def render_dot_action_node(self, node):
        label = self.render_node_label(node)
        node_name = self.render_node_name(node)
        return dict(name=node_name, label=label, shape="rectangle", rank=node.rank)



from typing import Dict
from h1st.exceptions.exception import GraphException
from h1st.trust.trustable import Trustable
# from h1st.schema.schema_validator import SchemaValidator
# from h1st.schema.schema_validation_result import SchemaValidationResult


class NodeContainable(Trustable):
    """
    Type of an instance to be added to a graph as a node must be the subclass of NodeContainable.
    .. code-block:: python
            :caption: Example of using NodeContainable

            import h1st.core as h1

            class GenerateWindowEvents(h1.NodeContainable):
                def call(self, command, inputs):
                    ...

            class MyGraph(h1.Graph):
                def __init__(self):
                    self
                        .start()
                        .add(GenerateWindowEvents())
                        .end()
        """
    def __init__(self):
        self._node = None
        
    @property
    def node(self) -> 'Node':
        if not self._node:
            from h1st.h1flow.h1flow import Graph 

            g = Graph()
            g.start()
            self._node = g.add(self)
            g.end()

        return self._node

    @property
    def graph(self) -> 'Graph':
        return self.node.graph

    def call(self, command: str, inputs: Dict) -> Dict:
        """
        Will be invoked by a node when executing a graph.

        Subclass of NodeContainable may override the call() function:
            import h1st.core as h1

            class MyClass(h1.NodeContainable)
                def call(self, command, inputs):
                    ...


        Or subclass may implement necessary functions which required for graph execution flows
        During executing, the function with name = value of command will be invoked
            class MyClass(h1.NodeContainable)                
                def predict(self, inputs):
                    # this function will be invoked if a graph execution is for predict:
                    # graph.predict(...)
                    # or graph.execute(command='predict', input_data=...)
                    ...
           
        :param command: to know which graph's execution flow (predict, train, ...) it is involving
        :inputs: input data to proceed accordingly to the flow

        :return: result as a dict
        """
        func = getattr(self, command)
        if not func:
            raise GraphException(f'class {self.__class__.__name__} must implement method "{command}')

        result = func(inputs)
        if not isinstance(result, dict):
            raise GraphException(f'output of {self.__class__.__name__} must be a dict')

        return result

    # def validate_node_output(self, input_data: dict=None, schema=None) -> SchemaValidationResult:
    #     """
    #     Subclass will implement this function to verify its output schema

    #     :param inputs: subclass will use this input data to call a specific function to get the result
    #     :schema: the schema to verify if the result conforming with
    #     """        
    #     return SchemaValidator().validate(input_data, schema)
